unit sak_util;


interface

uses sysutils, Windows, Registry, classes;

// this function by hou yg <yghou@yahoo.com>
function sak_IsIPAddress(ss : string):boolean;

// this function by Dmitry Bondarenko & Sergio Kessler
function sak_StrWord2Int( s: string;       //input text
                          wordNum: word;   //number of the word to convert
                          defVal: integer  //default value
                         ): integer;

function sak_GetBoundaryOutOfLine( s: string; position: integer): string;

// this function by Matjaz Bravc *)
function sak_GetInternetDate( Date: TDateTime):string;

function sak_MakeUniqueID( HostName: string): string;

function sak_GenerateBoundary: string;

function sak_GetTempFileName: string;

// this function by Chris Gnther, bugfixed by Sergio ;-)
function sak_QuotedPrintableDecode( FCurrentData : PChar ) : string;

// this function by Chris Gnther
function sak_Base64Decode( sOrg : PChar ) : string;

function sak_ConvertCharSet( Line : string ) : string;

function sak_GetMIMEType(const FileName : string) : string;

function sak_HexDigitToDecDigit( c: char): integer;

// this function by Chris Gnther
function sak_ExtractAddress( sAddress : string; WithDelimit : Boolean ) : string;

// this function by Chris Gnther
function sak_ExtractAlias( sAddress : string; WithDelimit : Boolean ) : string;

function sak_DelFromStr( sSource: string; sChar: Char ): string;

function sak_CleanUpAddresses( AddressList: string): TStringList;


implementation


// this function by hou yg <yghou@yahoo.com>
function sak_IsIPAddress(ss : string):boolean;
var s,s1 : string;
    ii,p : integer;
begin
   s1 := ss;
   result := false;
   for ii := 1 to 3 do begin
       p := pos('.',s1);
       if p = 0 then exit;
       s := Copy(s1,1,p-1);
       s1 := copy(s1,p+1,Length(s1));
       p := StrToIntDef(s,-1);
       if (p <= 0) or (p>=255) then exit;
   end;
   p := StrToIntDef(s1,-1);
   if (p <= 0) or (p>=255) then exit;
   result := true;
end;


// this function by Dmitry Bondarenko & Sergio Kessler
function sak_StrWord2Int( s: string;       //input text
                          wordNum: word;   //number of the word to convert
                          defVal: integer  //default value
                         ): integer;
var space_index, i: integer;
begin
     result := defVal;
     s := trim( s);

     for i := 1 to WordNum-1 do
     begin
          space_index := pos( ' ', s);
          if space_index = 0
             then exit
             else s := trim( copy( s, space_index, length(s)));
     end;

     space_index := pos( ' ', s);
     if space_index > 0 then s := trim( copy( s, 1, space_index));
     // here we delete the trailing words

     result := StrToIntDef( s, defVal);
end;

function sak_GetBoundaryOutOfLine( s: string; position: integer): string;
var x: integer;
begin
     result := trim( copy( s, position + length('BOUNDARY'), 72));
     delete( result, 1, 1); // remove the =
     result := trim( result);  // remove any leading spaces
     if result[1] = '"' then
     begin
          delete( result, 1, 1);
          x := pos( '"', result);      //sometimes there are a ; in the end
          if x > 0 then delete( result, x, 72);
     end;
     x := pos( ';', result);
     if x > 0 then delete( result, x, 72);
     result := '--' + trim( result);
end;


// this function by Matjaz Bravc
function sak_GetInternetDate( Date: TDateTime):string;
(* The date in RFC 822 conform string format *)

 function int2str(value:integer; len:byte):string;
  begin
       result := IntToStr( value);
       while length( result) < len do result := '0' + result;
  end;

 function GetTimeZoneBias:longint;
 (* The offset to UTC/GMT in minutes of the local time zone *)
 var tz_info: TTimeZoneInformation;
 begin
      case GetTimeZoneInformation(tz_info) of
           1: result := tz_info.StandardBias+tz_info.Bias;
           2: result := tz_info.DaylightBias+tz_info.Bias;
       else
           result := 0;
      end;
 end;

 function GetTimeZone:string;
 var bias: longint;
 begin
      bias := GetTimeZoneBias;
      if bias = 0
         then result := 'GMT'
         else if bias < 0
                 then result := '+' + int2str(abs(bias) div 60,2)+int2str(abs(bias) mod 60,2)
                 else if bias > 0 then result := '-' + int2str(bias div 60,2)+int2str(bias mod 60,2);
 end;

var d, m, y, w, h, mm, s, ms: word;
const
 WeekDays: array [1..7] of string[3] = ('Sun','Mon','Tue','Wed','Thu','Fri','Sat');
 Months: array [1..12] of string[3] = ('Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec');
begin
     DecodeDate( date, y, m, d);
     DecodeTime( date, h, mm, s, ms);
     w := DayOfWeek( date);
     Result := weekdays[w] + ', ' +
               inttostr(d) + ' ' +
               months[m] + ' ' +
               inttostr(y) + ' ' +
               int2str(h,2) + ':' +
               int2str(mm,2) + ':' +
               int2str(s,2) + ' ' +
               GetTimeZone;
end;

function sak_MakeUniqueID( HostName: string) : string;
var s: string;
    i : Integer;
begin
  Randomize;
  s := '';
  for i:=1 to 8 do
    s := Concat( s, chr( 97 + Random(20)));
  result := '<' + 'SAK.' +  formatDateTime( 'yyyy"."mm"."dd"."', date) +
            s + '@' + HostName + '>';
end;


function sak_GenerateBoundary : string;
var s: string;
    i : integer;
begin
     randomize;
     s := '';
     for i := 1 to 12 do
         s := s + chr( 97 + random( 25));
     result := '--_SAK_bound_' + s;
end;

function sak_GetTempFileName: string;
var tmpFileName: string;
begin
     SetLength( tmpFileName, 256);  // Extra byte for NULL
     GetTempPath(255, PChar( tmpFileName));
     GetTempFileName( PChar( tmpFileName), 'sak', 0, PChar( tmpFileName));
     SetLength( tmpFileName, StrLen( PChar( tmpFileName)));
     if FileExists( tmpFileName) then DeleteFile( PChar( tmpFileName));
     result := tmpFileName;
end;


// this function by Chris Gnther, bugfixed by Sergio ;-)
function sak_QuotedPrintableDecode( FCurrentData : PChar ): string;
{ This works if charset="iso-8859-1" ! }
var
   SourceIndex                : Integer;
   DecodedIndex               : Integer;
   Ch                         : Char;
   Code                       : Integer;
begin
   SourceIndex := 0;
   DecodedIndex := 0;
   if ( FCurrentData <> nil ) and ( FCurrentData^ <> #0 ) then
   begin
      while TRUE do
      begin
         Ch := FCurrentData[ SourceIndex ];
         if Ch = #0 then
            break;

         if Ch = '_' then
         begin
            FCurrentData[ DecodedIndex ] := ' ';
            Inc( SourceIndex );
            Inc( DecodedIndex );
         end
         else
            if Ch <> '=' then
            begin
               FCurrentData[ DecodedIndex ] := Ch;
               Inc( SourceIndex );
               Inc( DecodedIndex );
            end
            else
            begin
               Inc( SourceIndex );
               Ch := FCurrentData[ SourceIndex ];
               if (Ch = #13) or (Ch = #10) then
                 begin
                      Inc( SourceIndex );
                      Inc( SourceIndex );
                 end
               else
                 begin
                    Code := sak_HexDigitToDecDigit( Ch );
                    Inc( SourceIndex );
                    Ch := FCurrentData[ SourceIndex ];
                    if Ch <> #0 then
                    begin
                       Code := ( Code shl 4 ) + sak_HexDigitToDecDigit( Ch );
                       Inc( SourceIndex );
                    end;
                    FCurrentData[ DecodedIndex ] := Chr( Code );
                    Inc( DecodedIndex );
                 end;
            end;
      end;
      FCurrentData[ DecodedIndex ] := #0;
   end;
   Result := FCurrentData;
end;


// this function by Chris Gnther
function sak_Base64Decode( sOrg : PChar ) : string;
const Base64In: array[ 0..127 ] of Byte = (
      255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
      255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
      255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
      255, 255, 255, 255, 62, 255, 255, 255, 63, 52, 53, 54, 55,
      56, 57, 58, 59, 60, 61, 255, 255, 255, 64, 255, 255, 255,
      0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12,
      13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25,
      255, 255, 255, 255, 255, 255, 26, 27, 28, 29, 30, 31, 32,
      33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45,
      46, 47, 48, 49, 50, 51, 255, 255, 255, 255, 255 );
var
   ByteCount                  : Integer;
   SourceIndex                : Integer;
   DataOut                    : array[ 0..2 ] of Byte;
   DataIn0                    : Byte;
   DataIn1                    : Byte;
   DataIn2                    : Byte;
   DataIn3                    : Byte;
   DecodedIndex               : Integer;
begin
   DecodedIndex := 0;

   { Skip white spaces }
   SourceIndex := 0;

   while ( sOrg[ SourceIndex ] <> #0 ) and ( sOrg[ SourceIndex ] = ' ' ) do
      Inc( SourceIndex );

   { Decode until end of line. Replace coded chars byte decoded ones }
   while ( sOrg[ SourceIndex ] <> #0 ) and
      ( sOrg[ SourceIndex + 1 ] <> #0 ) and
      ( sOrg[ SourceIndex + 2 ] <> #0 ) and
      ( sOrg[ SourceIndex + 3 ] <> #0 ) and
      ( sOrg[ SourceIndex ] <> ' ' ) do
   begin
      DataIn0 := Base64In[ Byte( sOrg[ SourceIndex + 0 ] ) ];
      DataIn1 := Base64In[ Byte( sOrg[ SourceIndex + 1 ] ) ];
      DataIn2 := Base64In[ Byte( sOrg[ SourceIndex + 2 ] ) ];
      DataIn3 := Base64In[ Byte( sOrg[ SourceIndex + 3 ] ) ];

      DataOut[ 0 ] := ( DataIn0 and $3F ) shl 2 + ( DataIn1 and $30 ) shr 4;
      if DataIn2 <> $40 then
      begin
         DataOut[ 1 ] := ( DataIn1 and $0F ) shl 4 + ( DataIn2 and $3C ) shr 2;
         if DataIn3 <> $40 then
         begin
            DataOut[ 2 ] := ( DataIn2 and $03 ) shl 6 + ( DataIn3 and $3F );
            ByteCount := 3;
         end
         else
            ByteCount := 2;
      end
      else
         ByteCount := 1;

      { Replace coded characters (4) by decoded characters (up to 3) }
      Move( DataOut, sOrg[ DecodedIndex ], ByteCount );
      DecodedIndex := DecodedIndex + ByteCount;

      SourceIndex := SourceIndex + 4;
   end;

   sOrg[ DecodedIndex ] := #0;
   Result := sOrg;
end;


// this function by Chris Gnther and Sergio Kessler
function sak_ConvertCharSet( Line : string ) : string;
var
   strToDecode, RestBefore, RestAfter, strActual: string;
   iLast, iFirst, eFirst: integer;
   Encoding: char;
begin
   strActual := Line;

   iFirst := Pos( '=?', strActual );
   iLast := Pos( '?=', strActual );
   while (iFirst > 0) and (iLast > 0) do
   begin
        RestBefore := copy( strActual, 1, iFirst-1 );
        RestAfter := copy( strActual, iLast+2, length( strActual) );
        strToDecode := copy( strActual, iFirst+2, iLast-iFirst-2 );
        eFirst := pos( '?', strToDecode);
        Encoding := strToDecode[eFirst+1];
        strToDecode := copy( strToDecode, eFirst+3, length( strToDecode));
        if Encoding = 'Q' then
           strActual := RestBefore + sak_QuotedPrintableDecode( PChar( strToDecode) ) + RestAfter;
        if Encoding = 'B' then
           strActual := RestBefore + sak_Base64Decode( PChar( strToDecode) ) + RestAfter;

        iFirst := Pos( '=?', strActual );
        iLast := Pos( '?=', strActual );
   end;
   Result := strActual;
end;


function sak_GetMIMEType(const FileName : string) : string;
var
  Key : string;
begin
  Result:='';
  with TRegistry.Create do
  try
    RootKey:=HKEY_CLASSES_ROOT;
    Key:=ExtractFileExt(FileName);
    if KeyExists(Key) then
    begin
      OpenKey(Key,false);
      Result:=ReadString('Content Type');
      CloseKey;
    end;
  finally
    if Result='' then
      Result:='application/octet-stream';
    free;
  end;
end;


function sak_HexDigitToDecDigit( c: char): integer;
begin
   result := -1;
   if c in [ '0'..'9' ] then
      Result := Ord( c ) - Ord( '0' )
   else
      if c in [ 'A'..'F' ] then
         Result := Ord( c ) - Ord( 'A') + 10
      else
         if c in [ 'a'..'f' ] then
            Result := Ord( c ) - Ord( 'a') + 10;
end;


// this function by Chris Gnther
function sak_ExtractAddress( sAddress : string; WithDelimit : Boolean ) : string;
var
   iStart, iEnd               : integer;
begin
   iStart := Pos( '<', sAddress );
   iEnd := Pos( '>', sAddress );
   if ( iStart > 0 ) and ( iEnd > 0 ) then
      if WithDelimit then
         Result := '<' + copy( sAddress, iStart + 1, iEnd - iStart - 1 ) + '>'
      else
         Result := copy( sAddress, iStart + 1, iEnd - iStart - 1 )
   else
      Result := sAddress;
end;


// this function by Chris Gnther
function sak_ExtractAlias( sAddress : string; WithDelimit : Boolean ) : string;
var iStart: integer;
begin
   iStart := Pos( '<', sAddress );
   if ( iStart > 0 ) then
      if WithDelimit then
      begin
         Result := '"' + Trim( copy( sAddress, 1, iStart - 1 ) ) + '" '
      end
      else
      begin
         if iStart = 1 then
            Result := copy( sAddress, 2, length( sAddress ) - 2 )
         else
         begin
            Result := Trim( copy( sAddress, 1, iStart - 1 ) );
            Result := sak_DelFromStr( Result, '"' );
            Result := sak_DelFromStr( Result, '''' );
         end;
      end
   else
      Result := sAddress;
end;

function sak_DelFromStr( sSource: string; sChar: Char ): string;
// delete all the ocurrences of sChar in the string sSource
var i: Integer;
begin
     i := pos( sChar, sSource);
     while i > 0 do
     begin
          delete( sSource, i, 1);
          i := pos( sChar, sSource);
     end;
     Result := sSource;
end;


function sak_CleanUpAddresses( AddressList: string): TStringList;
var i: integer;
begin
     result := TStringList.Create;
     AddressList := trim( AddressList);
     if AddressList = '' then exit;

     i := pos( ',', AddressList);
     while i > 0 do
     begin                          // to various addresses
          result.Add( trim( copy( AddressList, 1, i-1)));
          delete( AddressList, 1, i);
          i := pos( ',', AddressList);
     end;
     result.Add( AddressList);            // the remaining address

     for i := 0 to result.Count-1 do
     begin
          result[i] := sak_ExtractAddress( result[i], false);
     end;
end;


end.
